global function InitTitanLimit

struct {
    array<entity> waitList
} file

array<string> HUDcreated = []
array<string> INFOsent = []

int maxTitans = 3
float maxPercentage = 0.6
int titanIncrease = 3
int scoreLimit = 650
int matchDuration = 15 * 60
int differenceLimit = 2

void function InitTitanLimit(){
    maxTitans = GetConVarInt( "TitanLimit" )
    maxPercentage = GetConVarFloat( "PlayerLimit" )
    differenceLimit = GetConVarInt( "DifferenceLimit" )
    titanIncrease = GetConVarInt( "TitanIncrease" )
    scoreLimit = GameMode_GetScoreLimit( "aitdm" )
    matchDuration = GameTime_TimeLeftSeconds()

    SetCallback_EarnMeterGoalEarned( EarnMeterMP_TitanEarned )
    AddSpawnCallback( "npc_titan", UpdateTitanAvailability )
    AddCallback_GameStateEnter( eGameState.Playing, Playing)
}

void function Playing(){

    print("[TITANLIMIT] Titan limit enabled")

    AddCallback_OnPlayerRespawned(OnPlayerSpawned)
    thread RunTitanfallAvailabiltyCheck()
    if (titanIncrease > 0){
        thread IncreaseLimit()
    }
}

void function OnPlayerSpawned(entity player){
    thread OnPlayerSpawnedThread(player)
}

void function OnPlayerSpawnedThread(entity player){
    print("[TITANLIMIT] Running on spawn")

    wait 8

    if ( IsValid(player) && INFOsent.find(player.GetPlayerName()) == -1){
        NSSendInfoMessageToPlayer( player, "This server has a titan limit. Status is displayed in the top right. Titans are granted to pilots in the order they were earned." );
        //Chat_ServerPrivateMessage( player, "This server has a titan limit, when your meter fills up, it will be saved. Titans are then granted to pilots in the order they were earned.", false)
        INFOsent.append(player.GetPlayerName())
    }

    wait 2

    if ( IsValid(player) && HUDcreated.find(player.GetPlayerName()) == -1){
        NSCreateStatusMessageOnPlayer( player, "[0/1]", "Titans deployed", "titanlimit");
        HUDcreated.append(player.GetPlayerName())
    }
}



float function MatchProgress(){ // Calculate and return the progress of the match as a value between zero and one, either based on score or time, whichever has progressed further
    float tempScoreProgress1 = 1.0 * GameRules_GetTeamScore2(TEAM_IMC) / scoreLimit
    float tempScoreProgress2 = 1.0 * GameRules_GetTeamScore2(TEAM_MILITIA) / scoreLimit
    float timeProgressLeft = 1.0 * GameTime_TimeLeftSeconds() / matchDuration
    if(tempScoreProgress1 == 0 && tempScoreProgress2 == 0){
        return 0.0001
    }
    if( tempScoreProgress1 < 1.0 - timeProgressLeft && tempScoreProgress1 < 1.0 - timeProgressLeft )
        return 1.0 - timeProgressLeft
    if(tempScoreProgress1 > tempScoreProgress2){
        return tempScoreProgress1
    }
    return tempScoreProgress2
}

void function IncreaseLimit(){
    float changeTreshold = 0.35 + ( 0.55 / titanIncrease )

    while(true){
        if( MatchProgress() >= changeTreshold){
            changeTreshold += 0.55 / titanIncrease
            if (maxTitans < maxTitans + titanIncrease){
                maxTitans += 1
                print( "[TITANLIMIT] Titan limit increased to " + maxTitans )
            }
        }
        if ( MatchProgress() > 0.90 ){
            return
        }
        wait 20
    }
}

void function UpdateHUD( int team ){
    print("[TITANLIMIT] HUD Updated")
    int titanCapacity = maxTitans
    int differenceLimitActive = 0

    if ((GetPlayerArray().len() / 2) * maxPercentage <= maxTitans){
        titanCapacity = (ceil(GetPlayerArray().len() * 0.5 * maxPercentage)).tointeger()
    }

    if (team == TEAM_IMC){
        int maxTitansImc = CountTeamTitans(TEAM_MILITIA) + differenceLimit

        if (maxTitansImc < maxTitans && maxTitansImc < titanCapacity){
            titanCapacity = maxTitansImc
            differenceLimitActive = 1
        }
        if (titanCapacity > maxTitans){
            titanCapacity = maxTitans
            differenceLimitActive = 0
        }

    }

    if (team == TEAM_MILITIA){
        int maxTitansMilitia = CountTeamTitans(TEAM_IMC) + differenceLimit

        if (maxTitansMilitia < maxTitans && maxTitansMilitia < titanCapacity){
            titanCapacity = maxTitansMilitia
            differenceLimitActive = 1
        }
        if (titanCapacity > maxTitans){
            titanCapacity = maxTitans
            differenceLimitActive = 0
        }
    }

    foreach (entity player in GetPlayerArray()){ // SendHudMessage(player, message, x_pos, y_pos, R, G, B, A, fade_in_time, hold_time, fade_out_time)
        if ( player.GetTeam() == team ) {
            if(CountTeamTitans(team) == titanCapacity && differenceLimitActive == 0){
                NSEditStatusMessageOnPlayer( player, "[" + CountTeamTitans(team) + "/" + titanCapacity + "]", "Titans deployed - max limit", "titanlimit" )
            }
            else if (CountTeamTitans(team) == titanCapacity && differenceLimitActive == 1){
                NSEditStatusMessageOnPlayer( player, "[" + CountTeamTitans(team) + "/" + titanCapacity + "]", "Titans deployed - diff limit", "titanlimit" )
            }
            else{
                NSEditStatusMessageOnPlayer( player, "[" + CountTeamTitans(team) + "/" + titanCapacity + "]", "Titans deployed", "titanlimit" )
            }
        }
    }
}

void function EarnMeterMP_TitanEarned( entity player )
{
    if ( CanPlayerHaveTitan( player ) )
    {
        if ( EarnMeterMP_IsTitanEarnGametype() ) // Check if pilot has ealready earned a titan, if so, simply reset their boost meter
        {
            SetTitanAvailable( player )
            //Remote_CallFunction_Replay( player, "ServerCallback_TitanReadyMessage" ) // broken for some reason
        }
        else
        {
            float oldRewardFrac = PlayerEarnMeter_GetRewardFrac( player )
            PlayerEarnMeter_Reset( player )
            PlayerEarnMeter_SetRewardFrac( player, oldRewardFrac )
            PlayerEarnMeter_EnableReward( player )

            if ( PlayerEarnMeter_GetRewardFrac( player ) != 0 )
                PlayerEarnMeter_EnableReward( player )
        }
    }
    else
    {
        float oldRewardFrac = PlayerEarnMeter_GetRewardFrac( player )
        PlayerEarnMeter_Reset( player )
        PlayerEarnMeter_SetRewardFrac( player, oldRewardFrac )
        PlayerEarnMeter_EnableReward( player )

        if ( PlayerEarnMeter_GetRewardFrac( player ) != 0 )
            PlayerEarnMeter_EnableReward( player )

        ClearTitanAvailable( player )

        //Chat_ServerPrivateMessage( player, "Your team has enough titans for now, your titan meter has been saved and you may earn boosts in the meantime.", false)

        foreach( entity p in file.waitList )
        {
            if ( p == player )
                return
        }
        print("[TITANLIMIT] " + player + " is waiting to drop their titan, added to queue" )

        file.waitList.append( player )
    } 
}

bool function CanPlayerHaveTitan( entity player ) {

    int militiaTitans = CountTeamTitans( TEAM_MILITIA )
    int imcTitans = CountTeamTitans( TEAM_IMC )
    bool militiaTf = true
    bool imcTf = true

    if ( militiaTitans > imcTitans ){
        imcTf = true
        if( militiaTitans - imcTitans >= differenceLimit){
            militiaTf = false
        }
    }
    else if ( imcTitans > militiaTitans ){
        militiaTf = true
        if( imcTitans - militiaTitans >= differenceLimit){
            imcTf = false
        }
    }
    if (maxTitans <= militiaTitans){
        militiaTf = false
    }
    if (maxTitans <= imcTitans){
        imcTf = false
    }

    if ( player.GetTeam() == TEAM_MILITIA )
        return militiaTf
    else{
        return imcTf
    }
    return true
}

int function CountTeamTitans( int team ) { // Count and return the number of titans on a team

    int titanCount = 0

    foreach ( entity titan in GetEntArrayByClass_Expensive("npc_titan") )
    {
        if ( titan.GetTeam() == team )
        {
            titanCount += 1
        }
    }
    foreach ( entity p in GetPlayerArray() )
    {
        if ( p.GetTeam() == team && p.IsTitan() )
        {
            titanCount += 1
        }
    }

    return titanCount
}

void function RunTitanfallAvailabiltyCheck(){

    wait 8

    foreach (entity player in GetPlayerArray()){
        NSSendInfoMessageToPlayer( player, "This server has a titan limit, when your meter fills up, it will be saved. Titans are then granted to pilots in the order they were earned." );
        //Chat_ServerPrivateMessage( player, "This server has a titan limit, when your meter fills up, it will be saved. Titans are then granted to pilots in the order they were earned.", false)
        INFOsent.append(player.GetPlayerName())
    }

    wait 2

    foreach (entity player in GetPlayerArray()){
        NSCreateStatusMessageOnPlayer( player, "[0/1]", "Titans deployed", "titanlimit");
        HUDcreated.append(player.GetPlayerName())
    }

    int imcTimer = 0
    int militiaTimer = 0
    int oldMilitiaTitans = 0
    int oldImcTitans = 0

    entity player
    for(;;) {
        int militiaTitans = CountTeamTitans( TEAM_MILITIA )
        int imcTitans = CountTeamTitans( TEAM_IMC )
        bool militiaTf = true
        bool imcTf = true
        array<entity> connected = GetPlayerArray()

        if ( militiaTitans > imcTitans ){
            imcTf = true
            if( militiaTitans - imcTitans >= differenceLimit){
                militiaTf = false
            }
        }
        else if ( imcTitans > militiaTitans ){
            militiaTf = true
            if( imcTitans - militiaTitans >= differenceLimit){
                imcTf = false
            }
        }
        if (maxTitans <= militiaTitans){
            militiaTf = false
        }
        if (maxTitans <= imcTitans){
            imcTf = false
        }
        if ((connected.len() / 2) * maxPercentage <= militiaTitans){
            militiaTf = false
        }
        if ((connected.len() / 2) * maxPercentage <= imcTitans){
            imcTf = false
        }

        if( imcTf && imcTimer == 0 ){                                        // Check IMC pilots
            for ( int i = file.waitList.len() - 1; i >= 0; i-- ) {
                player = file.waitList[i]
                if ( !IsValid( player ) ) {
                    file.waitList.remove( i )
                }
                else if ( player.GetTeam() == TEAM_IMC ) {  // This runs to give a player a titan
                    print("[TITANLIMIT] Titan awarded to " + player.GetPlayerName() )
                    NSSendAnnouncementMessageToPlayer( player, "TITANFALL READY", "Use it or lose it", <1,0,0>, 0, 1 )
                    file.waitList.remove( i )
                    PlayerEarnMeter_Reset( player )
                    player.p.earnMeterOverdriveFrac = 1.0
                    player.SetPlayerNetFloat( EARNMETER_EARNEDFRAC, 1.0 )
                    PlayerEarnMeter_SetOwnedFrac( player, 1.0 )
                    PlayerEarnMeter_SetRewardFrac( player, 1.0 )
                    imcTimer = 80
                    break
                }
            }
        }

        if( militiaTf && militiaTimer == 0 ){                                  // Check MILITIA pilots
            for ( int i = file.waitList.len() - 1; i >= 0; i-- ) {
                player = file.waitList[i]
                if ( !IsValid( player ) ) {
                    file.waitList.remove( i )
                }
                else if ( player.GetTeam() == TEAM_MILITIA ) {  // This runs to give a player a titan
                    print("[TITANLIMIT] Titan awarded to " + player.GetPlayerName() )
                    NSSendAnnouncementMessageToPlayer( player, "TITANFALL READY", "Use it or lose it", <1,0,0>, 0, 1 )
                    file.waitList.remove( i )
                    PlayerEarnMeter_Reset( player )
                    player.p.earnMeterOverdriveFrac = 1.0
                    player.SetPlayerNetFloat( EARNMETER_EARNEDFRAC, 1.0 )
                    PlayerEarnMeter_SetOwnedFrac( player, 1.0 )
                    PlayerEarnMeter_SetRewardFrac( player, 1.0 )
                    militiaTimer = 80
                    break
                }
            }
        }

        if (imcTimer > 0){
            imcTimer -= 1
        }
        if (militiaTimer > 0){
            militiaTimer -= 1
        }
        if ( oldImcTitans != imcTitans ){
            UpdateHUD( TEAM_IMC )
            oldImcTitans = imcTitans
        }
        if ( oldMilitiaTitans != militiaTitans ){
            UpdateHUD( TEAM_MILITIA )
            oldMilitiaTitans = militiaTitans
        }
        wait 0.05
    }
}


void function UpdateTitanAvailability(entity npcTitan)
{
    Assert( IsMultiplayer() )

    string settingsFile = GetNPCTitanSettingFile( npcTitan )

    // TitanLoadoutDef loadout = expect TitanLoadoutDef( GetTitanLoadoutForPrimary( primary ) )
}
